-- copyright benjamin bober, 2020
qualityoflife = qualityoflife or {};
local q = qualityoflife;

local defaultSourceSpeed = 160;
local defaultSourceSparks = 1;
local defaultSourceSparkFrequency = 0.01;

local defaultSparkGravity = 1;
local defaultSparkFriction = 1;
local defaultSparkLifetime = 2;
local defaultSparkSize = 1;

local sparkSources = {};
local function GetSparkSource()
	for id, source in ipairs(sparkSources) do
		if not source.busy then
			return source;
		end
	end

	local newSource = CreateFrame("Frame", "QoL_SparkSource" .. (table.getn(sparkSources) + 1), UIParent);
	newSource:SetWidth(1);
	newSource:SetHeight(1);
	newSource:Show();

	tinsert(sparkSources, newSource);
	return newSource;
end

local sparks = {};
local function GetSpark(source)
	for id, spark in ipairs(sparks) do
		if not spark.busy then
			return spark;
		end
	end

	local newSpark = CreateFrame("Frame", "QOL_Spark" .. (table.getn(sparks) + 1), UIParent);
	newSpark:SetBackdrop({bgFile = "Interface\\ChatFrame\\ChatFrameBackground"});

	tinsert(sparks, newSpark);
	return newSpark;
end

local function SparkBehaviour(spark, elapsed)
	spark.vx = spark.vx / spark.friction;
	spark.vy = spark.vy - (spark.gravity * elapsed);
	if spark.friction ~= 0 then
		spark.vy = spark.vy / spark.friction;
	end

	local point, relativeTo, relativePoint, xOfs, yOfs = spark:GetPoint();
	spark:ClearAllPoints();
	spark:SetPoint(point, relativeTo, relativePoint, xOfs + spark.vx, yOfs + spark.vy);
	spark:SetAlpha(spark:GetAlpha() - elapsed / spark.lifetime);

	if spark:GetAlpha() <= 0 then
		spark:SetScript("OnUpdate", nil);
		spark:Hide();
		spark.busy = false;
	end
end

local function AddSpark(source)
	local spark = GetSpark(source);
	spark.source = source;
	spark.busy = true;
	spark.vx = math.random();
	if math.random() > 0.5 then
		spark.vx = -spark.vx;
	end
	spark.vy = math.random();

	local r, g, b, a;
	if source.path[1]["rndColor"] then
		r = math.random();
		g = math.random();
		b = math.random();
		a = math.random();
	else
		r = source.path[1]["r"] or 1;
		g = source.path[1]["g"] or 1;
		b = source.path[1]["b"] or 1;
		a = source.path[1]["a"] or 1;
	end
	spark:SetBackdropColor(r, g, b, a);

	spark.gravity = source.gravity;
	spark.friction = source.friction;
	spark.lifetime = source.lifetime;

	spark:SetWidth(source.size);
	spark:SetHeight(source.size);
	spark:SetAlpha(1);

	local point, relativeTo, relativePoint, xOfs, yOfs = source:GetPoint();
	spark:SetParent(source);
	spark:SetPoint("BOTTOMLEFT", relativeTo, "BOTTOMLEFT", xOfs, yOfs);

	spark:SetScript("OnUpdate", function(self, elapsed)
		SparkBehaviour(self, elapsed);
	end);
	spark:Show();
end

local function PrepareSparkSourceSpeeds(source)
	source.v = source.path[1]["v"] or source.v or defaultSourceSpeed;
	if type(source.v) == "table" and not source.v["keys"] then
		local values, keys = q:Deepcopy(source.v), {};
		for key in pairs(source.v) do
			tinsert(keys, key);
		end
		table.sort(keys);

		source.v = {
			["values"] = values,
			["keys"] = keys
		};
	end
end

local function SetSparkSourcePosition(source, parent, x, y)
	parent = parent or UIParent;
	x, y = x or 0, y or 0;
	source:ClearAllPoints();
	source:SetParent(parent);
	source:SetPoint("BOTTOMLEFT", parent, "BOTTOMLEFT", x, y);
end

local function SparkSourceNextWaypoint(source, initial)
	if not initial then
		if source.path[1]["repeatAll"] then
			source.path = q:Deepcopy(source.pathCopy);
		elseif not source.path[1]["repeat"] then
			table.remove(source.path, 1);
		end
	end

	if table.getn(source.path) > 0 then
		local point, relativeTo, relativePoint, xOfs, yOfs = source:GetPoint();

		if source.path[1]["parent"] and _G[source.path[1]["parent"]] then
			relativeTo = _G[source.path[1]["parent"]];
		end
		xOfs = source.path[1]["x"] or xOfs or GetScreenWidth() / 2;
		yOfs = source.path[1]["y"] or yOfs or GetScreenHeight() / 2;
		SetSparkSourcePosition(source, relativeTo, xOfs, yOfs)

		local dx = source.path[1]["tx"] or xOfs;
		local dy = source.path[1]["ty"] or yOfs;
		local r = source.path[1]["r"] or 0;
		local phi1 = source.path[1]["phi1"] or 0;
		local phi2 = source.path[1]["phi2"] or 0;
		source.phi = phi2 - phi1;

		local txDist = q:MathRound(dx - xOfs, 2);
		local tyDist = q:MathRound(dy - yOfs, 2);
		local rDist = q:MathRound(r * source.phi, 2);
		local maxDist = math.max(math.abs(txDist), math.abs(tyDist), math.abs(rDist));

		source.distRatioX = txDist / maxDist;
		source.distRatioY  = tyDist / maxDist;
		source.distRatioW = rDist / maxDist;
		source.maxDist = maxDist;
		source.currentPhi = phi1 + math.pi / 2;
		source["X"], source["Y"], source["W"] = 0, 0, 0;
		PrepareSparkSourceSpeeds(source);
		source.delay = 0;

		source.friction = source.path[1]["friction"] or source.friction or defaultSparkFriction;
		source.gravity = source.path[1]["gravity"] or source.gravity or defaultSparkGravity;
		source.frequency = source.path[1]["frequency"] or source.frequency or defaultSourceSparkFrequency;
		source.lifetime = source.path[1]["lifetime"] or source.lifetime or defaultSparkLifetime;
		source.sparks = source.path[1]["sparks"] or source.sparks or defaultSourceSparks;
		source.size = source.path[1]["size"] or source.size or defaultSparkSize;

		if maxDist == math.abs(txDist) then
			source.refDist = "X";
		elseif maxDist == math.abs(tyDist) then
			source.refDist = "Y";
		else
			source.refDist = "W";
		end
	else
		q:HideSparkSource(source);
	end
end

local function GetSparkSourceSpeed(source)
	if type(source.v) == "table" then
		local waypointPerc = source[source.refDist] / source.maxDist;
		local lowPerc, highPerc;

		for _, key in ipairs(source.v["keys"]) do
			if waypointPerc >= key then
				lowPerc = key;
			else
				highPerc = key;
				break;
			end
		end

		if lowPerc and highPerc then
			local stepDist, stepProgress = highPerc - lowPerc, waypointPerc - lowPerc;
			return q:MathLerp(source.v["values"][lowPerc], source.v["values"][highPerc], stepProgress / stepDist);
		elseif lowPerc then
			return source.v["values"][lowPerc];
		end
	else
		return source.v;
	end
end

local deviation = 0.01;
local function SparkSourceBehaviour(source, elapsed)
	if source.path[1]["delay"] and (not source.delay or source.delay < source.path[1]["delay"]) then
		source.delay = source.delay or 0;
		source.delay = source.delay + elapsed;
	else
		if (source[source.refDist] < source.maxDist) then
			local point, relativeTo, relativePoint, xOfs, yOfs = source:GetPoint();
			local v = GetSparkSourceSpeed(source);
			local dx = v * source.distRatioX * elapsed;
			local dy = v * source.distRatioY * elapsed;
			local dr = v * source.distRatioW * elapsed;
			local drX = dr * math.cos(source.currentPhi);
			local drY = dr * math.sin(source.currentPhi);
			SetSparkSourcePosition(source, relativeTo, xOfs + dx + drX, yOfs + dy + drY);

			source["X"] = source["X"] + math.abs(dx);
			source["Y"] = source["Y"] + math.abs(dy);
			if source.phi ~= 0 then
				local adr = math.abs(dr);
				source["W"] = source["W"] + adr;
				source.currentPhi = source.currentPhi + source.phi * (adr / source.maxDist);
			end
		elseif source.path[1]["rest"] and (not source.rest or source.rest < source.path[1]["rest"]) then
			source.rest = source.rest or 0;
			source.rest = source.rest + elapsed;
		else
			SparkSourceNextWaypoint(source);
			return;
		end

		-- sparks
		source.lastSpark = source.lastSpark + elapsed
		if source.lastSpark >= source.frequency then
			for i = 1, source.sparks do
				AddSpark(source);
			end
			source.lastSpark = 0;
		end
	end
end

function q:ShowSparkSource(path)
	if path and type(path) == "table" and table.getn(path) > 1 then
		local source = GetSparkSource();
		source.busy = true;
		source.pathCopy = q:Deepcopy(path);
		source.path = q:Deepcopy(path);

		source.lastSpark = 0;
		source.friction = nil;
		source.gravity = nil;
		source.frequency = nil;
		source.lifetime = nil;
		source.sparks = nil;
		source.size = nil;

		SparkSourceNextWaypoint(source, true);

		source:SetScript("OnUpdate", function(self, elapsed)
			SparkSourceBehaviour(self, elapsed);
		end);
		source:Show();

		return source;
	end
end

function q:HideAllSparkSources()
	for id, source in ipairs(sparkSources) do
		q:HideSparkSource(source);
	end
end

function q:HideSparkSources(t)
	for _, source in ipairs(t) do
		q:HideSparkSource(source);
	end
end

function q:HideSparkSource(source)
	source:SetScript("OnUpdate", nil);
	source:Hide();
	source.busy = false;
end

function q:GetFrameOutlinePath(frame, repeatSequence)
	return {
		{
			["parent"] = frame:GetName(),
			["x"] = 0,
			["y"] = 0,
		},
		{
			["tx"] = 0,
			["ty"] = frame:GetHeight()
		},
		{
			["tx"] = frame:GetWidth(),
			["ty"] = frame:GetHeight()
		},
		{
			["tx"] = frame:GetWidth(),
			["ty"] = 0
		},
		{
			["tx"] = 0,
			["ty"] = 0,
			["repeatAll"] = repeatSequence
		}
	};
end